/*******************************************************************************
FILE : nullInterpreter.c

LAST MODIFIED : 16 January 2007

DESCRIPTION :
Just passes commands through without interpretation
==============================================================================*/
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is cmgui.
 *
 * The Initial Developer of the Original Code is
 * Auckland Uniservices Ltd, Auckland, New Zealand.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include <stdlib.h>
#include <regex.h>

enum Message_type
/*******************************************************************************
LAST MODIFIED : 31 May 1996

DESCRIPTION :
The different message types.
==============================================================================*/
{
	ERROR_MESSAGE,
	INFORMATION_MESSAGE,
	WARNING_MESSAGE
}; /* enum Message_type */

typedef void (*execute_command_function_type)(char *, void *, int *, int *);

typedef int (Interpreter_display_message_function)(enum Message_type message_type,
	char *format, ... );

struct Interpreter 
/*******************************************************************************
LAST MODIFIED : 16 January 2007

DESCRIPTION :
Internal data storage for this interpreter.
==============================================================================*/
{
	int number_of_variables;
	char **variable_names, **variable_values;
	regex_t *variable_assignment_regex;
	regex_t *variable_substitute_regex;
}; /* struct Interpreter */

void create_interpreter_(int argc, char **argv, const char *initial_comfile, 
	 struct Interpreter **interpreter, int *status)
/*******************************************************************************
LAST MODIFIED : 16 January 2007

DESCRIPTION:
Creates the interpreter for processing commands.
==============================================================================*/
{
  char error_string[100];
  int error_code, return_code;

  return_code = 1;

  if (*interpreter = (struct Interpreter *)malloc(sizeof (struct Interpreter)))
  {

 	  (*interpreter)->number_of_variables = 0; 
 	  (*interpreter)->variable_names = (char **)NULL; 
 	  (*interpreter)->variable_values = (char **)NULL; 

	  (*interpreter)->variable_assignment_regex = 
		  (regex_t *)malloc(sizeof(regex_t));
	  if (error_code = regcomp((*interpreter)->variable_assignment_regex,
	  	  "^[[:space:]]*[$]([[:alpha:]][[:alnum:]_]+)[[:space:]]*"
		  "[=][[:space:]]*(([[:digit:]]+)|[\"\']([^\"\']*)[\"\'])[[:space:]]*;?[[:space:]]*$"
			  , REG_EXTENDED))
	  {
		  regerror(error_code, (*interpreter)->variable_assignment_regex,
			  error_string, sizeof(error_string));
		  printf("REGEX error %s\n", error_string);
	  }

	  (*interpreter)->variable_substitute_regex = 
		  (regex_t *)malloc(sizeof(regex_t));
	  if (error_code = regcomp((*interpreter)->variable_substitute_regex,
	  	  "[$]([[:alpha:]][[:alnum:]_]+|[{][[:alpha:]][[:alnum:]_]+[}])"
			  , REG_EXTENDED))
	  {
		  regerror(error_code, (*interpreter)->variable_substitute_regex,
			  error_string, sizeof(error_string));
		  printf("REGEX error %s\n", error_string);
	  }
  }
  else
  {
	  *interpreter = (struct Interpreter *)NULL;
	  return_code = 0;
  }
  
  *status = return_code;
}

void destroy_interpreter_(struct Interpreter *interpreter, int *status)
/*******************************************************************************
LAST MODIFIED : 16 January 2007

DESCRIPTION:
==============================================================================*/
{
	int i;

	 if (interpreter)
	 {
		 regfree(interpreter->variable_assignment_regex);
		 free(interpreter->variable_assignment_regex);
		 regfree(interpreter->variable_substitute_regex);
		 free(interpreter->variable_substitute_regex);

		 if (interpreter->variable_names && 
			 interpreter->variable_values)
		 {
			 for (i = 0 ; i < interpreter->number_of_variables ; i++)
			 {
				 free(interpreter->variable_names[i]);
				 free(interpreter->variable_values[i]);
			 }
			 free(interpreter->variable_names);
			 free(interpreter->variable_values);
		 }
		 
		 free (interpreter);
		 *status = 1;
	 }
	 else
	 {
		 *status = 0;
	 }
}

void interpreter_set_display_message_function_(struct Interpreter *interpreter, 
	 Interpreter_display_message_function *function, int *status)
/*******************************************************************************
LAST MODIFIED : 16 January 2007

DESCRIPTION:
Sets the function that will be called whenever the Interpreter wants to report
information.
==============================================================================*/
{
	 *status = 1;
}

void redirect_interpreter_output_(struct Interpreter *interpreter, int *status)
/*******************************************************************************
LAST MODIFIED : 16 January 2007

DESCRIPTION:
This redirects the output from stdout to a pipe so that the handle_output
routine can write this to the command window.
==============================================================================*/
{
  *status = 1;
}

void interpret_command_(struct Interpreter *interpreter, char *command_string, 
	 void *user_data, int *quit,
  execute_command_function_type execute_command_function, int *status)
/*******************************************************************************
LAST MODIFIED : 16 January 2007

DESCRIPTION:
Takes a <command_string>, replaces simple variables and sends the command
back to the main program.
==============================================================================*/
{
	char *current_string, *new_name, *new_value, *temporary_string = NULL;
	int i, match, match_value_length, return_code;
	regmatch_t matches[5];

	if (interpreter)
	{
		if (0 == regexec(interpreter->variable_assignment_regex,
				command_string, sizeof(matches)/sizeof(regmatch_t), matches, 0))
		{
			if ((interpreter->variable_names = (char **)realloc(
					 interpreter->variable_names, 
					 (interpreter->number_of_variables + 1) * sizeof(char *)))
				&& (interpreter->variable_values = (char **)realloc(
					 interpreter->variable_values, 
					 (interpreter->number_of_variables + 1) * sizeof(char *))))
			{
				new_name = strndup(command_string + matches[1].rm_so,
					matches[1].rm_eo - matches[1].rm_so);

				if (-1 != matches[3].rm_so)
				{
					new_value = strndup(command_string + matches[3].rm_so,
						matches[3].rm_eo - matches[3].rm_so);
				}
				else if (-1 != matches[4].rm_so)
				{
					new_value = strndup(command_string + matches[4].rm_so,
						matches[4].rm_eo - matches[4].rm_so);
				}
				match = -1;
				for (i = 0 ; (-1 == match) &&
						  (i < interpreter->number_of_variables) ; i++)
				{
					if (!strcmp(interpreter->variable_names[i], new_name))
					{
						free(new_name);
						free(interpreter->variable_values[i]);
						interpreter->variable_values[i] = new_value;
						match = i;
						/* Won't worry about extra pointer on arrays as it
							will get used next time */
					}
				}
				if (-1 == match)
				{
					interpreter->variable_names[interpreter->number_of_variables] =
						new_name;
					interpreter->variable_values[interpreter->number_of_variables] =
						new_value;
#if defined (DEBUG)
					printf ("  new variable %s = %s\n", new_name, new_value);
#endif /* defined (DEBUG) */
					interpreter->number_of_variables++;
				}
			}
			return_code = 1;
		}
		else
		{
			current_string = command_string;
			return_code = 1;
			while (return_code && (0 == regexec(interpreter->variable_substitute_regex,
				current_string, sizeof(matches)/sizeof(regmatch_t), matches, 0)))
			{
#if defined (DEBUG)
				printf(" MATCH %d %d\n",
					matches[0].rm_so, matches[0].rm_eo);
#endif /* defined (DEBUG) */
				match = -1;
				for (i = 0 ; (match == -1) && (i < interpreter->number_of_variables) ; i++)
				{
#if defined (DEBUG)
					printf("Comparing %d characters %s  %s\n",
						strlen(interpreter->variable_names[i]),
						command_string + matches[0].rm_so + 1,
						interpreter->variable_names[i]);
#endif /* defined (DEBUG) */
					if (!strncmp(command_string + matches[0].rm_so + 1,
							interpreter->variable_names[i],
							strlen(interpreter->variable_names[i])))
					{
						match = i;
					}
				}
				if (-1 != match)
				{
					match_value_length = strlen(interpreter->variable_values[match]);
					temporary_string = (char *)realloc(temporary_string,
						strlen(current_string) + match_value_length);
					if (current_string == command_string)
					{
						/* Copying from the command_string */
						memcpy(temporary_string, current_string, matches[0].rm_so);
						memcpy(
							temporary_string + matches[0].rm_so + match_value_length,
							current_string + matches[0].rm_eo,
							strlen(current_string + matches[0].rm_eo) + 1);
						current_string = temporary_string;
					}
					else
					{
						/* The location of the string may have shifted */
						current_string = temporary_string;
						/* Shift the second part of the string first in case
							we are overlapping */
						memmove(temporary_string + matches[0].rm_so +
							match_value_length,
							temporary_string + matches[0].rm_eo,
							strlen(temporary_string + matches[0].rm_eo) + 1);
							
					}
					memcpy(current_string + matches[0].rm_so,
						interpreter->variable_values[match],
						match_value_length);
				}
				else
				{
					return_code = 0;
				}
			}
			if (return_code)
			{
				(execute_command_function)(current_string, 
					user_data, quit, &return_code);
			}
			else
			{
				printf("Error sustituting variable\n");
			}
			if (temporary_string)
			{
				free(temporary_string);
			}
		}
	}
	else
	{
		 return_code=0;
	}
	
	*status = return_code;
} /* interpret_command_ */

void interpreter_evaluate_string_(struct Interpreter *interpreter, 
	 char *expression, char **result, int *status)
/*******************************************************************************
LAST MODIFIED : 16 January 2007

DESCRIPTION:
NOT_IMPLEMENTED
==============================================================================*/
{
	*result = (char *)NULL;
	*status = 0;
} /* interpreter_evaluate_string_ */

void interpreter_set_string_(struct Interpreter *interpreter, 
	 char *variable_name, char *value, int *status)
/*******************************************************************************
LAST MODIFIED : 16 January 2007

DESCRIPTION:
NOT_IMPLEMENTED
==============================================================================*/
{
	*status = 0;
} /* interpreter_set_string_ */

void interpreter_set_pointer_(struct Interpreter *interpreter,
	 char *variable_name, char *class_name, void *value,int *status)
/*******************************************************************************
LAST MODIFIED : 16 January 2007

DESCRIPTION:
NOT_IMPLEMENTED
==============================================================================*/
{
	*status = 0;
} /* interpreter_set_string_ */
